/* main.c
 *
 * hashkill - a hash cracking tool
 * Copyright (C) 2010 Milen Rangelov <gat3way@gat3way.eu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <getopt.h>
#include <openssl/pem.h>
#include "loadfiles.h"
#include "hashinterface.h"
#include "plugins.h"
#include "threads.h"
#include "sessions.h"


/* Function prototypes */
static void usage(const char *progname);
static char* strupr(char* ioString);
static hash_stat parse_bruteforce_args(char *bruteargs);
int main(int argc, char *argv[]);




/* Display program usage */
static void usage(const char *progname)
{
    printf("\nUsage: %s [options]\n",progname);
    printf("\nOptions:\n-------------\n");
    printf("-P<plugin>, --plugin-info[=..] \t  Show plugins summary or detailed plugin information if <plugin> is provided\n");
    printf("-p <plugin>, --plugin          \t  Use plugin (default plugin is 'md5')\n");
    printf("-S<session>, --session-info[=..]  Display sessions summary/detailed info on <session>\n");
    printf("-s <session>, --session        \t  Load <session> (you may list sessions using -S)\n");
    printf("-f <hashfile>, --hashfile     \t  Load hashes from file \'hashfile\'\n");
    printf("-d <dictionary>, --dictionary \t  Enables dictionary mode (or hybrid if -b set), load dictionary \'dictionary\'\n");
    printf("-b<ruleset>, --brute[=...]    \t  Enables bruteforce mode (or hybrid if -d set), load ruleset  \'ruleset\' (default: \'1:8:lalphanum\')\n");
    printf("-t <prefix>, --prefix        \t  Bruteforce prefix\n");
    printf("-o <filename>, --outfile     \t  Write cracked hashes to <filename> upon completion\n");
    printf("-M, --markov-info[=...]      \t  Show available Markov stats files summary\n");
    printf("-m <statfile>, --markov    \t  Enable Markov attack, use <statfile>\n");
    printf("-n <threshold>,--markov-threshold Set Markov attack threshold\n");
    printf("-N <limit>, --markov-limit \t  Set Markov attack password length limit (<=10, default 8)\n");


    printf("\nRuleset format: \'start_len:end_len:predefined_set:additional_args\'\n\n");
    printf("start_len:  \t  generated plaintexts are equal in length or longer than that value\n"
	   "end_len:  \t  generated plaintexts are equal in length or shorter than that value\n"
	   "predefined_set:   one of \'alpha\', \'ualpha\', \'lalpha\', \'lalphanum\', \'ualphanum\', \'alphanum\', \'ascii\', \'none\'\n"
	   "additional_args:  additional characters, e.g \'+()\'\n\n"
	   "Example:\n\n -t hello -b1:2:lalpha:* - try all permutations of length 1-2, comprised of lower alphabetical chars and \'*\', starting with \'hello\'"
	   " (e.g hellogmark hello* hellojf, etc)\n");
    printf("\n");
}



/* Uppercase a string */
static char* strupr(char* ioString)
{
    int i;
    int theLength = (int)strlen(ioString);

    for(i=0; i<theLength; ++i) {ioString[i] = toupper(ioString[i]);}
    return ioString;
}


/* Parse brute-force arguments */
static hash_stat parse_bruteforce_args(char *bruteargs)
{
    char *arg1, *arg2, *arg3, *arg4, *argtmp;
    int tmp1=0, tmp2=0, counter, counter2, flag = 0;
    
    bzero(bruteforce_charset,255);
    arg3 = malloc(128);

    arg1 = strtok(bruteargs, ":");
    if (arg1) tmp1 = atoi(arg1);
    arg2 = strtok(NULL, ":");
    if (arg2) tmp2 = atoi(arg2);
    argtmp = strtok(NULL,":");
    bzero(arg3,128); // Make valgrind happy
    if (argtmp) strcpy(arg3, strupr(argtmp));
    arg4 = strtok(NULL,":");

    if ((tmp1 == 0) || (tmp2 == 0))
    {
	elog("Bad bruteforce parameters (start=end=%d)!\n",tmp1);
	if (arg3) free(arg3);
	return hash_err;
    }
    if (tmp1>tmp2)
    {
	elog("Bad bruteforce parameters, start>end (start=end=%d)!\n",tmp1);
	if (arg3) free(arg3);
	return hash_err;
    }
    else
    {
	bruteforce_start = tmp1;
	bruteforce_end = tmp2;
    }
    
    
    if (strcmp(arg3,"LALPHA") == 0)
    {
	strcpy(arg3,"abcdefghijklmnopqrstuvwxyz");
    }
    else if (strcmp(arg3,"UALPHA") == 0)
    {
	strcpy(arg3,"ABCDEFGHIJKLMNOPQRSTUVWXYZ");
    }
    else if (strcmp(arg3,"ALPHA") == 0)
    {
	strcpy(arg3,"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz");
    }
    else if (strcmp(arg3,"NUM") == 0)
    {
	strcpy(arg3,"0123456789");
    }
    else if (strcmp(arg3,"LALPHANUM") == 0)
    {
	strcpy(arg3,"0123456789abcdefghijklmnopqrstuvwxyz");
    }
    else if (strcmp(arg3,"UALPHANUM") == 0)
    {
	strcpy(arg3,"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ");
    }
    else if (strcmp(arg3,"ALPHANUM") == 0)
    {
	strcpy(arg3,"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz");
    }
    else if (strcmp(arg3,"ASCII") == 0)
    {
	strcpy(arg3,"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz!@#$%^&*()-_=+[]{}|\\\'\";:`~<>,./?");
    }
    else bzero(arg3,10);
    
    if (arg4) for (counter = 0; counter < (int)strlen(arg4); counter++)
    {
	flag = 0;
	for (counter2 = 0; counter2 < (int)strlen(arg3); counter2++) 
	    if (*(arg3+counter2) == *(arg4+counter)) flag = 1;
	if (flag ==0) 
	{
	    strncat(arg3, (arg4+counter), 1);
	    strcat(arg3,"");
	}
    }
    
    
    if (strlen(arg3)==0)
    {
	elog("Empty bruteforce charset!\n%s","");
	if (arg3) free(arg3);
	return hash_err;
    }
    else
    {
	strcpy(bruteforce_charset, arg3);
    }
    if (arg3) free(arg3);
    
    return hash_ok;

}


/* Program entrypoint */
int main(int argc, char *argv[])
{

    unsigned long totalmemory;
    unsigned long tempqueuesize;
    char *fvalue = NULL;
    char *dvalue = NULL;
    char *ovalue = NULL;
    char *Mvalue = NULL;
    char *mvalue = NULL;
    char *markovstat = NULL;
    int option,option_index;
    int pflag=0, bflag = 0, sflag = 0, mflag = 0;
    struct option long_options[] = 
    {
	{"plugin-info", 2, 0, 'P'},
	{"plugin", 1, 0, 'p'},
	{"session-info", 2, 0, 'S'},
	{"session", 1, 0, 's'},
	{"hashlist", 1, 0, 'f'},
	{"dictionary", 1, 0, 'd'},
	{"brute", 2, 0, 'b'},
	{"brute-prefix", 1, 0, 't'},
	{"outfile", 1, 0, 'o'},
	{"markov-info", 0, 0, 'M'},
	{"markov", 1, 0, 'm'},
	{"markov-threshold", 1, 0, 'n'},
	{"markov-limit", 1, 0, 'N'},
	{"help", 0, 0, 'h'},
	{0, 0, 0, 0}
    };

    /* initialize */
    printf("\n");
    hlog("Version %s\n", HASHKILL_VERSION);
    session_restore_flag = 0;
    hash_plugin_parse_hash = NULL;
    hash_plugin_check_hash = NULL;
    hash_list = NULL;
    cracked_list = NULL;
    attack_method = -1;
    bruteforce_prefix[0] = 0;
    markov_attack_init();
    attack_method = attack_method_markov;
    OpenSSL_add_all_algorithms();

    /* now store the command line parameters to be used by sessions module */
    session_put_commandline(argv);


    opterr = 0;
    while ((option = getopt_long(argc, argv, "p:f:d:P::b::t:T:S::s:o:N:n:M::m:h",long_options, &option_index)) != -1)
    switch (option)
    {
	case 'h':
	    usage(argv[0]);
	    exit(EXIT_SUCCESS);
	    break;
	

	case 's':
	    if (!optarg) {
		usage(argv[0]);
		exit(EXIT_FAILURE);
	    }
	    else {
		if (session_restore(optarg) == hash_err) exit(EXIT_FAILURE);
		sflag = 1;
		pflag = 1;
	    }
	break;


	case 'P':
	    if (!optarg) (void)print_plugins_summary("/usr/share/hashkill/plugins");
	    else print_plugin_detailed(optarg);
    	    exit(EXIT_SUCCESS);
	break;

	case 'b':
	    if (optarg) 
	    {
		if (parse_bruteforce_args(optarg) == hash_err)
		{
		    exit(EXIT_FAILURE);
		}
	    }
	    else
	    {
		char *tempopt = NULL;
		tempopt = malloc(100);
		strcpy(tempopt,"1:8:lalphanum:");
		if (parse_bruteforce_args(tempopt) == hash_err)
		{
		    free(tempopt);
		    exit(EXIT_FAILURE);
		}
		free(tempopt);
	    }
    	    bflag = 1;
	break;
	
	case 'm':
	    if (optarg) 
	    {
		if (markov_load_statfile(optarg) == hash_err)
		{
		    exit(EXIT_FAILURE);
		}
		else
		{
		    markovstat = alloca(64);
		    strcpy(markovstat,optarg);
		    attack_method = attack_method_markov;
		}
	    }
	    else
	    {
		usage(argv[0]);
		exit(EXIT_FAILURE);
	    }
    	    mflag = 1;
	break;


	case 't':
	    if (optarg) strcpy(bruteforce_prefix, optarg);
	    break;


	case 'S':
	    if (!optarg) (void)print_sessions_summary();
	    else print_session_detailed(optarg);
	    exit(EXIT_SUCCESS);
	break;


	
	case 'p':
	    set_current_plugin(optarg);
            if (load_plugin()==hash_err)
	    {
    	        elog("Cannot load plugin (%s)\n",optarg);
    		exit(EXIT_FAILURE);
    	    }
	    pflag=1;
	break;

        case 'n':
	    markov_max_len = atoi(optarg);
	break;

        case 'N':
	    markov_threshold = atoi(optarg);
	break;


        
        case 'M':
	    markov_print_statfiles();
	    exit(EXIT_SUCCESS);
	break;
        
	case 'f':
	    fvalue = optarg;
	break;
        
	case 'd':
	    dvalue = optarg;
	break;

	case 'o':
	    ovalue = optarg;
	break;


        case '?':
	    if ((optopt == 'f') || (optopt == 'P') || (optopt == 'd'))
	    {
		fprintf(stderr, "Option -%c requires an argument.\n", optopt);
    	    }
/*		else
    		{
    		    fprintf (stderr, "Unknown option character '%c'.\n", optopt);
        	    return 1;
		}*/
	
	default:
	break;
    }


    if (pflag==0) 
    {
        set_current_plugin("md5");
        if (load_plugin()==hash_err)
        {
    	    elog("Cannot load default plugin (%s)\n","md5");
    	    exit(EXIT_FAILURE);
    	}
    }

    if (fvalue) 
    {
	strncpy(hashlist_file,fvalue,254);
	hashlist_file[254] = 0;
	(void)load_hashes_file(fvalue);
    }



    /* Do we have argument (hash)? */
    if (argv[optind]) 
    {
	if (load_single_hash(argv[optind]) == hash_err)
	{
	    if (!fvalue) 
	    {
		elog("Cannot load hash: %s\n",argv[optind]);
		exit(EXIT_FAILURE);
	    }
	}
    }


    /* Hashes num */
    hashes_count = get_hashes_num();

    /* Dictionary? */
    if ((dvalue)&&(!bflag))
    {
	attack_method = attack_method_dictionary;
	strcpy(hash_dictionary_file, dvalue);
    }

    /* Bruteforce? */
    else if ((bflag)&&(!dvalue))
    {
	attack_method = attack_method_simple_bruteforce;
    }

    /* hybrid? */
    else if ( (bflag) && (dvalue))
    {
	attack_method = attack_method_hybrid;
	strcpy(hash_dictionary_file, dvalue);

    }
    

    if ((attack_method != attack_method_simple_bruteforce) && ((attack_method != attack_method_markov)) && (attack_method != attack_method_dictionary) && (attack_method != attack_method_hybrid))
    {
	usage(argv[0]);
	exit(EXIT_FAILURE);
    }
// queue/threads sizing
    if (get_hashes_num() == 0)
    {
	elog("No hashes loaded!%s (try --help)\n","");
	exit(EXIT_FAILURE);
    }
    
    totalmemory = hash_get_total_memory();
    hash_num_threads = hash_num_cpu()*4;
    tempqueuesize = (totalmemory * 1024 * 1024) / (4 * hash_num_threads * 512);


    if (tempqueuesize > HASHKILL_MAXQUEUESIZE) tempqueuesize = HASHKILL_MAXQUEUESIZE;
    
    hlog("Threads queue size: %lu plaintexts/thread\n", tempqueuesize);
    if (spawn_threads(hash_num_threads, (unsigned int)tempqueuesize) == hash_err) exit(EXIT_FAILURE);

    if (attack_method == attack_method_markov) 
    {
	
	if (!markovstat) markov_load_statfile("rockyou");
	(void)main_thread_markov();
    }


    if (attack_method == attack_method_dictionary) 
    {
	(void)main_thread_dictionary();
    }
    
    if (attack_method == attack_method_simple_bruteforce) 
    {
	(void)main_thread_bruteforce();
    }

    if (attack_method == attack_method_hybrid) 
    {
	(void)main_thread_hybrid();
    }


    if (get_cracked_num() > 0) print_cracked_list();    
    session_unlink_file();
    if (ovalue)
    {
	print_cracked_list_to_file(ovalue);
    }
    thread_attack_cleanup();

    cleanup_lists();

    hlog("Bye bye :)\n\n%s","");
    return 0;
}
